package com.codigoparallevar.minicards.parts.logic;

import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.Rect;
import android.util.Log;

import com.codigoparallevar.minicards.PartInstantiator;
import com.codigoparallevar.minicards.ScrolledCanvas;
import com.codigoparallevar.minicards.parts.connectors.RoundOutputConnector;
import com.codigoparallevar.minicards.types.Moveable;
import com.codigoparallevar.minicards.types.Part;
import com.codigoparallevar.minicards.types.PartConnection;
import com.codigoparallevar.minicards.types.PartGrid;
import com.codigoparallevar.minicards.types.functional.Tuple2;
import com.codigoparallevar.minicards.types.connectors.input.InputConnector;
import com.codigoparallevar.minicards.types.connectors.output.OutputConnector;
import com.codigoparallevar.minicards.types.wireData.Signal;
import com.codigoparallevar.minicards.types.wireData.WireDataType;
import com.codigoparallevar.minicards.utils.Serializations;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.UUID;

public class Ticker implements Part {

    private static final int DEFAULT_SIDE_SIZE = 100;
    private final String _id;
    private final PartGrid _partGrid;
    private int _left;
    private int _top;
    private int _right;
    private int _bottom;
    private List<OutputConnector> _outputConnectors;
    private final RoundOutputConnector _signalOutputConnector;
    private final long SLEEP_TIME = 1000;
    private Thread _thread = null;

    private Ticker(String id, PartGrid partGrid, int left, int top, int right, int bottom) {
        _id = id;
        _partGrid = partGrid;
        _left = left;
        _top = top;
        _right = right;
        _bottom = bottom;

        // Create connectors
        _signalOutputConnector = new RoundOutputConnector(
                this,
                _partGrid,
                getOutputConnectorCenterX(), getOutputConnectorCenterY(),
                getOutputConnectRadius());

        _outputConnectors = new LinkedList<>();
        _outputConnectors.add(_signalOutputConnector);
    }


    public Ticker(PartGrid partGrid, int left, int top, int right, int bottom) {
        this(UUID.randomUUID().toString(), partGrid, left, top, right, bottom);
    }


    @Override
    public void moveEnd(int x, int y) {
        final int width = _right - _left;
        final int height = _bottom - _top;

        _left = x - width / 2;
        _right = _left + width;

        _top = y - height / 2;
        _bottom = _top + height;

        _signalOutputConnector.updatePosition(
                getOutputConnectorCenterX(),
                getOutputConnectorCenterY());
    }

    @Override
    public void drop(int x, int y) {
        moveEnd(x, y);
    }

    @Override
    public boolean containsPoint(int x, int y) {
        return (x >= get_left()) && (x <= get_right())
                && (y >= get_top()) && (y <= get_bottom());
    }

    @Override
    public Moveable getMoveable() {
        return this;
    }

    @Override
    public void unlink() {
        pause();

        for (InputConnector input : getInputConnectors()) {
            input.unlink();
        }
    }

    @Override
    public void draw(ScrolledCanvas canvas, boolean devMode) {
        if (devMode){
            drawConnector(canvas);
            drawWires(canvas, devMode);

            Paint paint = new Paint(Paint.ANTI_ALIAS_FLAG);
            paint.setColor(Color.GRAY);

            canvas.drawRect(
                    new Rect(_left, _top,
                            _right, _bottom),
                    paint);


            // Craw a little clock
            Paint clockPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
            clockPaint.setStyle(Paint.Style.STROKE);
            clockPaint.setColor(Color.YELLOW);
            clockPaint.setStrokeWidth(2f);

            int halfX = (_left + _right) / 2;
            int halfY = (_top + _bottom) / 2;
            int radius = (_bottom - _top) / 2;

            canvas.drawCircle(halfX, halfY,
                    radius * .8f, clockPaint);

            canvas.drawLine(halfX, halfY - radius * 0.2f,
                    halfX, halfY - radius * 0.8f,
                    clockPaint);
        }
    }

    private void drawConnector(ScrolledCanvas canvas) {
        Paint connectorPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
        connectorPaint.setColor(Color.RED);

        canvas.drawCircle(getOutputConnectorCenterX(),
                getOutputConnectorCenterY(),
                getOutputConnectRadius(),
                connectorPaint);
    }

    private void drawWires(ScrolledCanvas canvas, boolean devMode) {
        for (OutputConnector outputConnector : _outputConnectors){
            outputConnector.drawWires(canvas, devMode);
        }
    }

    @Override
    public int get_left() {
        return _left;
    }

    @Override
    public int get_right() {
        return _right;
    }

    @Override
    public int get_top() {
        return _top;
    }

    @Override
    public int get_bottom() {
        return _bottom;
    }

    @Override
    public void touched() {
        // Just ignore it, as it's a logic component
    }

    @Override
    public List<InputConnector> getInputConnectors() {
        return Collections.emptyList();
    }

    @Override
    public List<OutputConnector> getOutputConnectors() {
        return _outputConnectors;
    }

    @Override
    public JSONObject serialize() throws JSONException {
        JSONObject serialized = new JSONObject();

        serialized.put("id", _id);
        serialized.put("left", _left);
        serialized.put("top", _top);
        serialized.put("right", _right);
        serialized.put("bottom", _bottom);
        serialized.put("on_signal_output_connector",
                Serializations.serialize(serializeConnectionEndpoints()));

        return serialized;
    }

    public static Tuple2<Part, List<PartConnection>> deserialize(PartGrid partGrid, JSONObject data) throws JSONException {
        String id = data.getString("id");
        int left = data.getInt("left");
        int top = data.getInt("top");
        int right = data.getInt("right");
        int bottom = data.getInt("bottom");

        Ticker ticker = new Ticker(id, partGrid, left, top, right, bottom);

        List<PartConnection> connections = new LinkedList<>();

        JSONArray connectorOuts = data.getJSONArray("on_signal_output_connector");
        for (int i = 0; i < connectorOuts.length(); i++){
            connections.add(PartConnection.deserialize(
                    ticker._signalOutputConnector,
                    connectorOuts.getJSONObject(i)));
        }

        return new Tuple2<Part, List<PartConnection>>(ticker, connections);
    }

    private List<Map<String, String>> serializeConnectionEndpoints() {
        List<Map<String, String>> serializedData = new LinkedList<>();

        for (Tuple2<String, String> endpoint : _signalOutputConnector.getConnectionEndpoints()){
            serializedData.add(PartConnection.serialize(endpoint.item1, endpoint.item2));
        }

        return serializedData;
    }

    @Override
    public void send(InputConnector roundInputConnector, WireDataType signal) {
        // @TODO: REMOVE THE NEED FOR THIS
    }

    @Override
    public String get_id() {
        return _id;
    }

    @Override
    public InputConnector getConnectorWithId(String inputConnectorId) {
        return null;
    }

    @Override
    public String getConnectorId(InputConnector inputConnector) {
        return null;
    }

    @Override
    public void resume() {
        _thread = new Thread(new Runnable() {
            @Override
            public void run() {
                while (Ticker.this._thread == Thread.currentThread()) {
                    Ticker.this._signalOutputConnector.send(new Signal());
                    _partGrid.update();

                    try {
                        Thread.sleep(SLEEP_TIME);
                    } catch (InterruptedException e) {
                        Log.e("Minicards Ticker", "Wait failed", e);
                        Ticker.this._thread = null;
                        return;
                    }
                }
            }
        });

        _thread.start();
    }

    @Override
    public void pause() {
        _thread = null;
    }


    private int getOutputConnectorCenterX() {
        return (_left + _right) / 2;
    }

    private int getOutputConnectorCenterY() {
        return _bottom;
    }

    private int getOutputConnectRadius() {
        return (_right - _left) / 2;
    }

    public static PartInstantiator getInstantiator() {
        final int halfSideSize = DEFAULT_SIDE_SIZE / 2;
        return new PartInstantiator() {
            @Override
            protected Part instantiate(PartGrid grid, Tuple2<Integer, Integer> center) {
                return new Ticker(grid,
                        center.item1 - halfSideSize, center.item2 - halfSideSize,
                        center.item1 + halfSideSize, center.item2 + halfSideSize);
            }
        };
    }
}
